/*
 * Copyright (c) 2018 by Intinor AB. All rights reserved.
 *
 * This file is part of "libRaptorQ".
 *
 * libRaptorQ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * libRaptorQ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * and a copy of the GNU Lesser General Public License
 * along with libRaptorQ.  If not, see <http://www.gnu.org/licenses/>.
 */


#pragma once

#include "RaptorQ/v1/common.hpp"
#include "RaptorQ/v1/Octet.hpp"
#include <stdlib.h>
#include <cstdlib>
#include <iostream>
#include <Eigen/Dense>
#include <Eigen/Sparse>

namespace RaptorQ__v1 {
namespace Impl {

using DenseMtx = Eigen::Matrix<Octet, Eigen::Dynamic, Eigen::Dynamic,
                                                            Eigen::RowMajor>;

// This class represents a read-write view of an Matrix
class RAPTORQ_LOCAL Matrix_Block
{
friend class Matrix;
public:
    uint8_t& operator() (uint32_t row, uint32_t col)
    {
        assert(row < block_rows);
        assert(col < block_cols);
        return block_data[index(row, col)];
    }

    int32_t rows() const { return static_cast<int32_t>(this->block_rows); }
    int32_t cols() const { return static_cast<int32_t>(this->block_cols); }

    DenseMtx to_eigen()
    {
        DenseMtx ret = DenseMtx(block_rows, block_cols);

        for (uint32_t i = 0; i < block_rows; i++) {
            for (uint32_t j = 0; j < block_cols; j++) {
                ret(i, j) = block_data[index(i, j)];
            }
        }

        return ret;
    }

    void set_zero()
    {
        for (int32_t i = 0; i < rows(); i++)
        {
            memset(&block_data[index(i, 0)], 0, cols());
        }
    }

private:
    Matrix_Block(uint8_t *data, uint32_t start_row, uint32_t start_column,
                                uint32_t rows, uint32_t columns, int32_t stride)
    {
        block_rows = rows;
        block_cols = columns;
        block_step = stride;
        block_data = data + index(start_row, start_column);
    }

    uint8_t *block_data {NULL}; // Owned by Matrix
    uint32_t block_rows {0};
    uint32_t block_cols {0};
    uint32_t block_step {0};

    uint32_t index(uint32_t row, uint32_t col)
    {
        return block_step * row + col;
    }
};

}
}
